/*
	Code released to public domain

	The obligatory "Please don't sue me" paragraph:

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
	IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
	OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
	OTHER DEALINGS IN THE SOFTWARE.

*/

#include <Windows.h>
#include <stdio.h>
#include <conio.h>
#include <assert.h>


// constants
//------------------------------------------------------------------------------------------


const BYTE cUseMaxVolume = 1;

// size of cNotes array
#define NOTES 13

const char* cNotes[] = {"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C"};

const BYTE cMaxVolume = 0x7F;

typedef enum 
{
	// reference: https://midi.org/specifications/item/table-2-expanded-messages-list-status-bytes
	eNoteOff = 0x80,
	eNoteOn = 0x90,
	eAfterTouch = 0xA0,
	eControlChange = 0xB0,
	eProgramChange = 0xC0,
	eChannelAfterTouch = 0xD0,
	ePitchWheel = 0xE0,
	eSystemMessage = 0xF0

}eMidiMessageType;


// globals
//------------------------------------------------------------------------------------------
HMIDIOUT OutputMidiDevice;



// functions
//------------------------------------------------------------------------------------------




BYTE GetNoteOctave(BYTE iNoteID)
{
	//Notes are numbered from C-1 (0x00) all the way up to G9 (0x7F)
	//http://cdn.instructables.com/FWX/NBXG/H4AFZWE7/FWXNBXGH4AFZWE7.MEDIUM.jpg

	return (iNoteID / 12);
}

BYTE GetNote(BYTE iNoteID)
{
	return (iNoteID % 12);
}

MMRESULT SendToOutput(BYTE iStatus, BYTE iNoteNumber, BYTE iVelocity)
{
	

	BYTE tSpare = 0;
	BYTE tMidiSecondByteData;

	if(cUseMaxVolume)
	{
		// if your keyboard is pressure sensitive and very stiff like mine, and you want to always hear notes
		// at max volume (127), set cUseMaxVolume to a nonzero value.
		tMidiSecondByteData = cMaxVolume;
	}

	else
	{

		tMidiSecondByteData = iVelocity;
	}

	BYTE tMidiFirstByteData = iNoteNumber;
	BYTE tMidiStatus = iStatus;

	//						Low, High
	WORD tHighMidiMessage = MAKEWORD(tMidiSecondByteData, tSpare);
	WORD tLowMidiMessage = MAKEWORD(iStatus, tMidiFirstByteData);

	DWORD_PTR tMidiMessage = MAKELONG(tLowMidiMessage, tHighMidiMessage);

	MMRESULT qResult;

	qResult = midiOutShortMsg(OutputMidiDevice, tMidiMessage);

	return qResult;
}


void PrintMidiDevices()
{
	UINT tNumberOfInputDevices = midiInGetNumDevs();
	UINT tNumberOfOutputDevices = midiOutGetNumDevs();
	UINT tMidiDeviceIndex = 0;

	printf("Listing midi devices...\n");
	printf("\n");

	printf("%d input device(s) found.\n", tNumberOfInputDevices);

	for(tMidiDeviceIndex = 0 ; tMidiDeviceIndex < tNumberOfInputDevices; tMidiDeviceIndex++)
	{
		UINT tDeviceToUse = tMidiDeviceIndex;
		MMRESULT tResult = 0;

		MIDIINCAPS tMidiDeviceInfo;
		UINT tDeviceStructSize = sizeof(MIDIINCAPS);

		memset(&tMidiDeviceInfo, 0, sizeof(MIDIINCAPS));
		
		tResult = midiInGetDevCaps(tDeviceToUse, &tMidiDeviceInfo, tDeviceStructSize);

		printf("%d: %s\n", tDeviceToUse, tMidiDeviceInfo.szPname);
	}

	printf("\n");

	printf("%d output device(s) found.\n", tNumberOfOutputDevices);

	for(tMidiDeviceIndex = 0; tMidiDeviceIndex < tNumberOfOutputDevices; tMidiDeviceIndex++)
	{
		UINT tDeviceToUse = tMidiDeviceIndex;
		MMRESULT tResult = 0;

		MIDIOUTCAPS tMidiDeviceInfo;
		UINT tDeviceStructSize = sizeof(MIDIOUTCAPS);

		memset(&tMidiDeviceInfo, 0, sizeof(MIDIOUTCAPS));
		
		tResult = midiOutGetDevCaps(tDeviceToUse, &tMidiDeviceInfo, tDeviceStructSize);

		printf("%d: %s\n", tDeviceToUse, tMidiDeviceInfo.szPname);
	}

	printf("\n");
}

// Please note that DWORD_PTR inputs to this function are NOT pointers.
void CALLBACK MidiInCallback(HMIDIIN hMidiIn, UINT wMsg, DWORD_PTR dwInstance, DWORD_PTR dwParam1, DWORD_PTR dwParam2)
{
	if(wMsg = MIM_DATA)
	{
		//https://msdn.microsoft.com/en-us/library/vs/alm/dd757284%28v=vs.85%29.aspx
		DWORD_PTR tMidiMessage = dwParam1;
		DWORD_PTR tTimestamp = dwParam2;

		WORD tHighMidiMessage = HIWORD(tMidiMessage);
		WORD tLowMidiMessage = LOWORD(tMidiMessage);

		BYTE tSpare = HIBYTE(tHighMidiMessage);
		BYTE tMidiSecondByteData = LOBYTE(tHighMidiMessage);

		BYTE tMidiFirstByteData = HIBYTE(tLowMidiMessage);
		BYTE tMidiStatus = LOBYTE(tLowMidiMessage);

		eMidiMessageType tMidiMessageType = (tMidiStatus & 0xF0);

		switch(tMidiMessageType)
		{

			case eNoteOff:
			{
				//printf("got Note Off Message!\n");
				BYTE tNoteNumber = tMidiFirstByteData; // 0-127
				BYTE tVelocity = tMidiSecondByteData; // 0-127

				BYTE tNoteOctave = GetNoteOctave(tNoteNumber);
				BYTE tNote = GetNote(tNoteNumber);

				char* tNoteName = NULL;

				if(tNote >= NOTES)
				{
					assert(tNote >= NOTES);
					break;
				}

				tNoteName = cNotes[tNote];

				//printf("Raw: %x Octave: %x Note: %x (%s), Volume: %x Time: %x\n", tNoteNumber, tNoteOctave, tNote, tNoteName, tVelocity, tTimestamp);

				SendToOutput(tMidiStatus, tNoteNumber, tVelocity, tTimestamp);

				break;
			}

			case eNoteOn:
			{
				//printf("got Note On Message!\n");
				BYTE tNoteNumber = tMidiFirstByteData; // 0-127
				BYTE tVelocity = tMidiSecondByteData; // 0-127

				BYTE tNoteOctave = GetNoteOctave(tNoteNumber);
				BYTE tNote = GetNote(tNoteNumber);

				char* tNoteName = NULL;

				if(tNote >= NOTES)
				{
					assert(tNote >= NOTES);
					break;
				}

				tNoteName = cNotes[tNote];

				printf("Raw: %x Octave: %x Note: %x (%s), Volume: %x Time: %x\n", tNoteNumber, tNoteOctave, tNote, tNoteName, tVelocity, tTimestamp);

				SendToOutput(tMidiStatus, tNoteNumber, tVelocity, tTimestamp);

				break;
			}

			// unsupported messages (for debugging)
			case eAfterTouch:
				printf("After touch message.\n");
				break;

			case eControlChange:
				printf("Control change message.\n");
				break;

			case eProgramChange:
				printf("Program change message.\n");
				break;

			case eChannelAfterTouch:
				printf("Channel after touch message.\n");
				break;

			case ePitchWheel:
				printf("Pitch wheel message.\n");
				break;

			case eSystemMessage:
				printf("System message.\n");
				break;

			default:
				printf("Unknown message %x.\n", tMidiMessageType);
				break;
		}

		//printf("got MIM_DATA message!\n");

	}
	
	
}


int main()
{

	MMRESULT tResult = 0;

	DWORD_PTR tInstanceData = 0; // note: not a pointer

	DWORD_PTR tInputStreamCallback = (DWORD_PTR) (&MidiInCallback);

	HMIDIIN tInputMidiDevice;	

	// see output of PrintMidiDevices() and edit these numbers below to match your MIDI device setup.
	const UINT tInputDeviceToUse = 0;
	const UINT tOutputDeviceToUse = 0;

	memset(&tInputMidiDevice, 0, sizeof(HMIDIIN));

	PrintMidiDevices();	

	printf("Connecting keyboard on input channel %d to (GS Wavetable) Synth on output channel %d\n", tInputDeviceToUse, tOutputDeviceToUse);
	printf("Ready. Press keys on your midi keyboard to hear them played over your speakers.\n");
	printf("Or press any key to quit.\n");
	printf("------------------------------------------------------------------------------------\n");
	printf("MIDI Message Log:\n\n");


	tResult = midiInOpen(&tInputMidiDevice, tInputDeviceToUse, MidiInCallback, tInstanceData, CALLBACK_FUNCTION);

	tResult = midiOutOpen(&OutputMidiDevice, tOutputDeviceToUse, NULL, tInstanceData, CALLBACK_NULL);

	// need to do this to get keypresses from your midi keyboard.
	tResult = midiInStart(tInputMidiDevice);
	
	

	// while we're waiting for any key, the callback will fire and re-route your keypresses on the midi keyboard
	// to GS Wavetable Synth.
	//
	// This means that you'll hear the keys you're pressing on your midi keyboard played through your speakers, 
	// using the (default) acoustic grand instrument.
	_getch();

	tResult = midiInStop(tInputMidiDevice);
	tResult = midiInClose(tInputMidiDevice);
	tResult = midiOutClose(OutputMidiDevice);


	return 0;
}

